"""
Deprecated implementation of basic and digest auth.

Look to auth_basic and auth_digest instead.
"""

import warnings

import cherrypy


# import late to avoid deprecation warning on import
httpauth = None


def import_httpauth():
    """
    Late-import cherrypy.lib.httpauth into globals.
    Then replace this function so it does nothing.
    """
    globals().update(
        httpauth=__import__('cherrypy.lib.httpauth').lib.httpauth,
        import_httpauth=lambda: None
    )


def check_auth(users, encrypt=None, realm=None):
    """If an authorization header contains credentials, return True or False.
    """
    import_httpauth()
    msg = (
        '`basic_auth` and `digest_auth` tools are deprecated. Use '
        '`auth_basic` and `auth_digest` instead.'
    )
    warnings.warn(msg, DeprecationWarning)
    request = cherrypy.serving.request
    if 'authorization' in request.headers:
        # make sure the provided credentials are correctly set
        ah = httpauth.parseAuthorization(request.headers['authorization'])
        if ah is None:
            raise cherrypy.HTTPError(400, 'Bad Request')

        if not encrypt:
            encrypt = httpauth.DIGEST_AUTH_ENCODERS[httpauth.MD5]

        if hasattr(users, '__call__'):
            try:
                # backward compatibility
                users = users()  # expect it to return a dictionary

                if not isinstance(users, dict):
                    raise ValueError(
                        'Authentication users must be a dictionary')

                # fetch the user password
                password = users.get(ah['username'], None)
            except TypeError:
                # returns a password (encrypted or clear text)
                password = users(ah['username'])
        else:
            if not isinstance(users, dict):
                raise ValueError('Authentication users must be a dictionary')

            # fetch the user password
            password = users.get(ah['username'], None)

        # validate the authorization by re-computing it here
        # and compare it with what the user-agent provided
        if httpauth.checkResponse(ah, password, method=request.method,
                                  encrypt=encrypt, realm=realm):
            request.login = ah['username']
            return True

        request.login = False
    return False


def basic_auth(realm, users, encrypt=None, debug=False):
    """If auth fails, raise 401 with a basic authentication header.

    realm
        A string containing the authentication realm.

    users
        A dict of the form: {username: password} or a callable returning
        a dict.

    encrypt
        callable used to encrypt the password returned from the user-agent.
        if None it defaults to a md5 encryption.

    """
    import_httpauth()
    if check_auth(users, encrypt):
        if debug:
            cherrypy.log('Auth successful', 'TOOLS.BASIC_AUTH')
        return

    # inform the user-agent this path is protected
    cherrypy.serving.response.headers[
        'www-authenticate'] = httpauth.basicAuth(realm)

    raise cherrypy.HTTPError(
        401, 'You are not authorized to access that resource')


def digest_auth(realm, users, debug=False):
    """If auth fails, raise 401 with a digest authentication header.

    realm
        A string containing the authentication realm.
    users
        A dict of the form: {username: password} or a callable returning
        a dict.
    """
    import_httpauth()
    if check_auth(users, realm=realm):
        if debug:
            cherrypy.log('Auth successful', 'TOOLS.DIGEST_AUTH')
        return

    # inform the user-agent this path is protected
    cherrypy.serving.response.headers[
        'www-authenticate'] = httpauth.digestAuth(realm)

    raise cherrypy.HTTPError(
        401, 'You are not authorized to access that resource')



import cherrypy
from cherrypy.test import helper

script_names = ['', '/path/to/myapp']


class ProxyTest(helper.CPWebCase):

    @staticmethod
    def setup_server():

        # Set up site
        cherrypy.config.update({
            'tools.proxy.on': True,
            'tools.proxy.base': 'www.mydomain.test',
        })

        # Set up application

        class Root:

            def __init__(self, sn):
                # Calculate a URL outside of any requests.
                self.thisnewpage = cherrypy.url(
                    '/this/new/page', script_name=sn)

            @cherrypy.expose
            def pageurl(self):
                return self.thisnewpage

            @cherrypy.expose
            def index(self):
                raise cherrypy.HTTPRedirect('dummy')

            @cherrypy.expose
            def remoteip(self):
                return cherrypy.request.remote.ip

            @cherrypy.expose
            @cherrypy.config(**{
                'tools.proxy.local': 'X-Host',
                'tools.trailing_slash.extra': True,
            })
            def xhost(self):
                raise cherrypy.HTTPRedirect('blah')

            @cherrypy.expose
            def base(self):
                return cherrypy.request.base

            @cherrypy.expose
            @cherrypy.config(**{'tools.proxy.scheme': 'X-Forwarded-Ssl'})
            def ssl(self):
                return cherrypy.request.base

            @cherrypy.expose
            def newurl(self):
                return ("Browse to <a href='%s'>this page</a>."
                        % cherrypy.url('/this/new/page'))

            @cherrypy.expose
            @cherrypy.config(**{
                'tools.proxy.base': None,
            })
            def base_no_base(self):
                return cherrypy.request.base

        for sn in script_names:
            cherrypy.tree.mount(Root(sn), sn)

    def testProxy(self):
        self.getPage('/')
        self.assertHeader('Location',
                          '%s://www.mydomain.test%s/dummy' %
                          (self.scheme, self.prefix()))

        # Test X-Forwarded-Host (Apache 1.3.33+ and Apache 2)
        self.getPage(
            '/', headers=[('X-Forwarded-Host', 'http://www.example.test')])
        self.assertHeader('Location', 'http://www.example.test/dummy')
        self.getPage('/', headers=[('X-Forwarded-Host', 'www.example.test')])
        self.assertHeader('Location', '%s://www.example.test/dummy' %
                          self.scheme)
        # Test multiple X-Forwarded-Host headers
        self.getPage('/', headers=[
            ('X-Forwarded-Host', 'http://www.example.test, www.cherrypy.test'),
        ])
        self.assertHeader('Location', 'http://www.example.test/dummy')

        # Test X-Forwarded-For (Apache2)
        self.getPage('/remoteip',
                     headers=[('X-Forwarded-For', '192.168.0.20')])
        self.assertBody('192.168.0.20')
        # Fix bug #1268
        self.getPage('/remoteip',
                     headers=[
                         ('X-Forwarded-For', '67.15.36.43, 192.168.0.20')
                     ])
        self.assertBody('67.15.36.43')

        # Test X-Host (lighttpd; see https://trac.lighttpd.net/trac/ticket/418)
        self.getPage('/xhost', headers=[('X-Host', 'www.example.test')])
        self.assertHeader('Location', '%s://www.example.test/blah' %
                          self.scheme)

        # Test X-Forwarded-Proto (lighttpd)
        self.getPage('/base', headers=[('X-Forwarded-Proto', 'https')])
        self.assertBody('https://www.mydomain.test')

        # Test X-Forwarded-Ssl (webfaction?)
        self.getPage('/ssl', headers=[('X-Forwarded-Ssl', 'on')])
        self.assertBody('https://www.mydomain.test')

        # Test cherrypy.url()
        for sn in script_names:
            # Test the value inside requests
            self.getPage(sn + '/newurl')
            self.assertBody(
                "Browse to <a href='%s://www.mydomain.test" % self.scheme +
                sn + "/this/new/page'>this page</a>.")
            self.getPage(sn + '/newurl', headers=[('X-Forwarded-Host',
                                                   'http://www.example.test')])
            self.assertBody("Browse to <a href='http://www.example.test" +
                            sn + "/this/new/page'>this page</a>.")

            # Test the value outside requests
            port = ''
            if self.scheme == 'http' and self.PORT != 80:
                port = ':%s' % self.PORT
            elif self.scheme == 'https' and self.PORT != 443:
                port = ':%s' % self.PORT
            host = self.HOST
            if host in ('0.0.0.0', '::'):
                import socket
                host = socket.gethostname()
            expected = ('%s://%s%s%s/this/new/page'
                        % (self.scheme, host, port, sn))
            self.getPage(sn + '/pageurl')
            self.assertBody(expected)

        # Test trailing slash (see
        # https://github.com/cherrypy/cherrypy/issues/562).
        self.getPage('/xhost/', headers=[('X-Host', 'www.example.test')])
        self.assertHeader('Location', '%s://www.example.test/xhost'
                          % self.scheme)

    def test_no_base_port_in_host(self):
        """
        If no base is indicated, and the host header is used to resolve
        the base, it should rely on the host header for the port also.
        """
        headers = {'Host': 'localhost:8080'}.items()
        self.getPage('/base_no_base', headers=headers)
        self.assertBody('http://localhost:8080')



# This file is part of CherryPy <http://www.cherrypy.org/>
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab:fileencoding=utf-8
"""HTTP Basic Authentication tool.

This module provides a CherryPy 3.x tool which implements
the server-side of HTTP Basic Access Authentication, as described in
:rfc:`2617`.

Example usage, using the built-in checkpassword_dict function which uses a dict
as the credentials store::

    userpassdict = {'bird' : 'bebop', 'ornette' : 'wayout'}
    checkpassword = cherrypy.lib.auth_basic.checkpassword_dict(userpassdict)
    basic_auth = {'tools.auth_basic.on': True,
                  'tools.auth_basic.realm': 'earth',
                  'tools.auth_basic.checkpassword': checkpassword,
                  'tools.auth_basic.accept_charset': 'UTF-8',
    }
    app_config = { '/' : basic_auth }

"""

import binascii
import unicodedata

import cherrypy
from cherrypy._cpcompat import base64_decode, ntob, ntou, tonative


__author__ = 'visteya'
__date__ = 'April 2009'


def checkpassword_dict(user_password_dict):
    """Returns a checkpassword function which checks credentials
    against a dictionary of the form: {username : password}.

    If you want a simple dictionary-based authentication scheme, use
    checkpassword_dict(my_credentials_dict) as the value for the
    checkpassword argument to basic_auth().
    """
    def checkpassword(realm, user, password):
        p = user_password_dict.get(user)
        return p and p == password or False

    return checkpassword


def basic_auth(realm, checkpassword, debug=False, accept_charset='utf-8'):
    """A CherryPy tool which hooks at before_handler to perform
    HTTP Basic Access Authentication, as specified in :rfc:`2617`
    and :rfc:`7617`.

    If the request has an 'authorization' header with a 'Basic' scheme, this
    tool attempts to authenticate the credentials supplied in that header.  If
    the request has no 'authorization' header, or if it does but the scheme is
    not 'Basic', or if authentication fails, the tool sends a 401 response with
    a 'WWW-Authenticate' Basic header.

    realm
        A string containing the authentication realm.

    checkpassword
        A callable which checks the authentication credentials.
        Its signature is checkpassword(realm, username, password). where
        username and password are the values obtained from the request's
        'authorization' header.  If authentication succeeds, checkpassword
        returns True, else it returns False.

    """

    fallback_charset = 'ISO-8859-1'

    if '"' in realm:
        raise ValueError('Realm cannot contain the " (quote) character.')
    request = cherrypy.serving.request

    auth_header = request.headers.get('authorization')
    if auth_header is not None:
        # split() error, base64.decodestring() error
        msg = 'Bad Request'
        with cherrypy.HTTPError.handle((ValueError, binascii.Error), 400, msg):
            scheme, params = auth_header.split(' ', 1)
            if scheme.lower() == 'basic':
                decoded_params = base64_decode(params)
                decoded_params = ntob(decoded_params)

                last_err = None
                for charset in (accept_charset, fallback_charset):
                    try:
                        decoded_params = tonative(decoded_params, charset)
                        break
                    except ValueError as ve:
                        last_err = ve
                else:
                    raise last_err

                decoded_params = ntou(decoded_params)
                decoded_params = unicodedata.normalize('NFC', decoded_params)
                decoded_params = tonative(decoded_params)
                username, password = decoded_params.split(':', 1)
                if checkpassword(realm, username, password):
                    if debug:
                        cherrypy.log('Auth succeeded', 'TOOLS.AUTH_BASIC')
                    request.login = username
                    return  # successful authentication

    charset = accept_charset.upper()
    charset_declaration = (
        (', charset="%s"' % charset)
        if charset != fallback_charset
        else ''
    )
    # Respond with 401 status and a WWW-Authenticate header
    cherrypy.serving.response.headers['www-authenticate'] = (
        'Basic realm="%s"%s' % (realm, charset_declaration)
    )
    raise cherrypy.HTTPError(
        401, 'You are not authorized to access that resource')



import os

import cherrypy
from cherrypy.test import helper

curdir = os.path.join(os.getcwd(), os.path.dirname(__file__))


class VirtualHostTest(helper.CPWebCase):

    @staticmethod
    def setup_server():
        class Root:

            @cherrypy.expose
            def index(self):
                return 'Hello, world'

            @cherrypy.expose
            def dom4(self):
                return 'Under construction'

            @cherrypy.expose
            def method(self, value):
                return 'You sent %s' % value

        class VHost:

            def __init__(self, sitename):
                self.sitename = sitename

            @cherrypy.expose
            def index(self):
                return 'Welcome to %s' % self.sitename

            @cherrypy.expose
            def vmethod(self, value):
                return 'You sent %s' % value

            @cherrypy.expose
            def url(self):
                return cherrypy.url('nextpage')

            # Test static as a handler (section must NOT include vhost prefix)
            static = cherrypy.tools.staticdir.handler(
                section='/static', dir=curdir)

        root = Root()
        root.mydom2 = VHost('Domain 2')
        root.mydom3 = VHost('Domain 3')
        hostmap = {'www.mydom2.com': '/mydom2',
                   'www.mydom3.com': '/mydom3',
                   'www.mydom4.com': '/dom4',
                   }
        cherrypy.tree.mount(root, config={
            '/': {
                'request.dispatch': cherrypy.dispatch.VirtualHost(**hostmap)
            },
            # Test static in config (section must include vhost prefix)
            '/mydom2/static2': {
                'tools.staticdir.on': True,
                'tools.staticdir.root': curdir,
                'tools.staticdir.dir': 'static',
                'tools.staticdir.index': 'index.html',
            },
        })

    def testVirtualHost(self):
        self.getPage('/', [('Host', 'www.mydom1.com')])
        self.assertBody('Hello, world')
        self.getPage('/mydom2/', [('Host', 'www.mydom1.com')])
        self.assertBody('Welcome to Domain 2')

        self.getPage('/', [('Host', 'www.mydom2.com')])
        self.assertBody('Welcome to Domain 2')
        self.getPage('/', [('Host', 'www.mydom3.com')])
        self.assertBody('Welcome to Domain 3')
        self.getPage('/', [('Host', 'www.mydom4.com')])
        self.assertBody('Under construction')

        # Test GET, POST, and positional params
        self.getPage('/method?value=root')
        self.assertBody('You sent root')
        self.getPage('/vmethod?value=dom2+GET', [('Host', 'www.mydom2.com')])
        self.assertBody('You sent dom2 GET')
        self.getPage('/vmethod', [('Host', 'www.mydom3.com')], method='POST',
                     body='value=dom3+POST')
        self.assertBody('You sent dom3 POST')
        self.getPage('/vmethod/pos', [('Host', 'www.mydom3.com')])
        self.assertBody('You sent pos')

        # Test that cherrypy.url uses the browser url, not the virtual url
        self.getPage('/url', [('Host', 'www.mydom2.com')])
        self.assertBody('%s://www.mydom2.com/nextpage' % self.scheme)

    def test_VHost_plus_Static(self):
        # Test static as a handler
        self.getPage('/static/style.css', [('Host', 'www.mydom2.com')])
        self.assertStatus('200 OK')
        self.assertHeader('Content-Type', 'text/css;charset=utf-8')

        # Test static in config
        self.getPage('/static2/dirback.jpg', [('Host', 'www.mydom2.com')])
        self.assertStatus('200 OK')
        self.assertHeaderIn('Content-Type', ['image/jpeg', 'image/pjpeg'])

        # Test static config with "index" arg
        self.getPage('/static2/', [('Host', 'www.mydom2.com')])
        self.assertStatus('200 OK')
        self.assertBody('Hello, world\r\n')
        # Since tools.trailing_slash is on by default, this should redirect
        self.getPage('/static2', [('Host', 'www.mydom2.com')])
        self.assertStatus(301)
